package at.ac.sbg.zgis.configuration;

import android.os.Environment;
import android.util.Log;
import android.util.SparseArray;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Objects;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import at.ac.sbg.zgis.survey.Survey;
import at.ac.sbg.zgis.survey.SurveyItem;
import at.ac.sbg.zgis.survey.surveyitem.RadioButtonGroup;
import at.ac.sbg.zgis.survey.surveyitem.Slider;
import at.ac.sbg.zgis.survey.surveyitem.TextInput;


public class Configuration {

    public static final String CONFIG_FILE_NAME = "config.xml"; // "config_version_long.xml", "config_one_element.xml"
    public static final String PLATFORM = "Platform";
    public static final String API_KEY = "API_key";
    public static final String QUESTION = "Question";
    private static final String CATEGORY = "Category";
    private static final String ANSWERS = "Answers";
    private static final String ANSWER = "Answer";
    private static final String MAXIMUM = "Maximum";
    private static final String RADIO_BUTTON_GROUP = "RadioButtonGroup";
    private static final String SLIDER = "Slider";
    private static final String TAG = Configuration.class.getSimpleName();
    private static final String SENSOR = "Sensor";
    private static final String SURVEY_ITEM = "SurveyItem";
    private static final String ACTIVE = "Active";
    private static final String TEXT_INPUT = "TextInput";
    private static final String USER_INTERFACE = "UserInterface";
    private static final String SURVEY_IS_ACTIVATED = "SurveyIsActivated";
    private static final String REAL_TIME_VISUALIZATION_IS_ACTIVATED = "RealTimeVisualizationIsActivated";
    private static final String MARKER_BUTTON_IS_ACTIVATED = "MarkerButtonIsActivated";


    private final List<HashMap<String, String>> sensors = new ArrayList<>();
    private Survey survey = null;

    private Boolean surveyIsActivated = true;
    private Boolean realTimeVisualizationIsActivated = true;
    private Boolean markerButtonIsActivated = true;


    /**
     * Reads the XML file and assigns all settings
     */
    public void importConfiguration() throws ParserConfigurationException, IOException, SAXException, NoSuchFieldException {

        File appDirectory = new File(Environment.getExternalStorageDirectory(), Constants.extDir);
        File file = new File(appDirectory, CONFIG_FILE_NAME);

        DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
        DocumentBuilder db = dbf.newDocumentBuilder();
        Document dom = db.parse(file);
        dom.getDocumentElement().normalize();

        readSensors(dom);

        readSurvey(dom);

        readUserInterfaceSettings(dom);


        Log.d(TAG, "Reading config file finished.");
    }


    /**
     * reads the survey information from the XML
     */
    private void readSurvey(Document dom) throws NoSuchFieldException {
        Log.d(TAG, "SurveyXML - start parsing");

        SparseArray<SurveyItem> surveyItemSparseArray = new SparseArray<>();


        // survey items
        NodeList surveyItemsNL = dom.getElementsByTagName(SURVEY_ITEM);
        for (int k = 0; k < surveyItemsNL.getLength(); k++) {

            // initialize properties of survey item, will be filled later
            String category = "";
            String question = "";
            String maximum = "";
            ArrayList<String> answers = new ArrayList<>();

            // survey properties
            Node surveyProperty = surveyItemsNL.item(k);
            if (surveyProperty.getNodeType() == Node.ELEMENT_NODE) {
                Element surveyPropertyElement = (Element) surveyProperty;

                // category
                category = surveyPropertyElement.getElementsByTagName(CATEGORY).item(0).getTextContent();
                Log.d(TAG, "SurveyXML - category " + category);

                // question
                question = surveyPropertyElement.getElementsByTagName(QUESTION).item(0).getTextContent();
                Log.d(TAG, "SurveyXML - question " + question);

                // maximum
                NodeList maximumNL = surveyPropertyElement.getElementsByTagName(MAXIMUM);
                if (maximumNL.getLength() != 0) {
                    maximum = maximumNL.item(0).getTextContent();
                    Log.d(TAG, "SurveyXML - maximum " + maximum);
                }

                // answers
                NodeList answerContainerNodeList = surveyPropertyElement.getElementsByTagName(ANSWERS);
                if (answerContainerNodeList.getLength() != 0) {
                    NodeList answersNodeList = ((Element) answerContainerNodeList.item(0)).getElementsByTagName(ANSWER);

                    if (answersNodeList.getLength() != 0) {
                        for (int m = 0; m < answersNodeList.getLength(); m++) {

                            String answer = answersNodeList.item(m).getTextContent();
                            Log.d(TAG, "SurveyXML - answer " + answer);
                            answers.add(answer);
                        }
                    }
                }
            }

            // create survey item objects
            SurveyItem surveyItem;

            if (Objects.equals(category, RADIO_BUTTON_GROUP)) {
                Log.d(TAG, "SurveyXML - create RadioButtonGroup Item");

                surveyItem = new RadioButtonGroup(question, answers);

            } else if (Objects.equals(category, SLIDER)) {
                Log.d(TAG, "SurveyXML - create Slider Item");

                int maximumInt = Integer.parseInt(maximum);
                surveyItem = new Slider(question, maximumInt);

            } else if (Objects.equals(category, TEXT_INPUT)) {
                Log.d(TAG, "SurveyXML - create Text Input Item");

                surveyItem = new TextInput(question);

            } else {
                Log.e(TAG, "SurveyXML - category not found");

                throw new NoSuchFieldException("Unknown survey item.");
            }

            int index = surveyItemSparseArray.size();
            surveyItemSparseArray.append(index, surveyItem);

        } // survey item - end

        this.survey = new Survey(surveyItemSparseArray);


        Log.d(TAG, "SurveyXML - Survey object created: " + survey);
    }


    /**
     * Reads the sensor settings from the XML
     */
    private void readSensors(Document dom) {

        // loop through sensors
        NodeList sensorNodes = dom.getElementsByTagName(SENSOR);
        for (int i = 0; i < sensorNodes.getLength(); i++) {

            // for storing infos about sensor
            HashMap<String, String> sensorHashMap = new HashMap<>();

            if (sensorNodes.item(i).getNodeType() == Node.ELEMENT_NODE) {
                Element sensorPropertyElement = (Element) sensorNodes.item(i);

                // platform
                String platform = sensorPropertyElement.getElementsByTagName(PLATFORM).item(0).getTextContent();
                sensorHashMap.put(PLATFORM, platform);

                // TODO: 24.09.18 Make Boolean
                // active
                String active = sensorPropertyElement.getElementsByTagName(ACTIVE).item(0).getTextContent();
                sensorHashMap.put(ACTIVE, active);

                // API Key
                NodeList apiKeyNL = sensorPropertyElement.getElementsByTagName(API_KEY);
                if (apiKeyNL.getLength() != 0) {
                    String apiKey = apiKeyNL.item(0).getTextContent();
                    sensorHashMap.put(API_KEY, apiKey);
                }
            }
            sensors.add(sensorHashMap);
        }
        Log.d(TAG, "Sensor settings read.");
    }


    private void readUserInterfaceSettings(Document dom) {

        // loop through sensors
        NodeList uiNodes = dom.getElementsByTagName(USER_INTERFACE);
        for (int i = 0; i < uiNodes.getLength(); i++) {
            Element uiPropertyElement = (Element) uiNodes.item(i);

            String surveyIsActivatedString = uiPropertyElement.getElementsByTagName(SURVEY_IS_ACTIVATED).item(0).getTextContent();
            surveyIsActivated = Boolean.parseBoolean(surveyIsActivatedString);

            String realTimeVisualizationIsActivatedString = uiPropertyElement.getElementsByTagName(REAL_TIME_VISUALIZATION_IS_ACTIVATED).item(0).getTextContent();
            realTimeVisualizationIsActivated = Boolean.parseBoolean(realTimeVisualizationIsActivatedString);

            String markerButtonIsActivatedString = uiPropertyElement.getElementsByTagName(MARKER_BUTTON_IS_ACTIVATED).item(0).getTextContent();
            markerButtonIsActivated = Boolean.parseBoolean(markerButtonIsActivatedString);
        }

    }

    /**
     * List with hashmaps that describe the sensor properties
     */
    public List<HashMap<String, String>> getSensors() {
        return Collections.unmodifiableList(sensors);
    }

    public Survey getSurvey() {
        return survey;
    }

    public Boolean isSurveyActivated() {
        return surveyIsActivated;
    }

    public Boolean isRealTimeVisualizationActivated() {
        return realTimeVisualizationIsActivated;
    }

    public Boolean isMarkerButtonActivated() {
        return markerButtonIsActivated;
    }

}