package at.ac.sbg.zgis.recording;

import android.content.ContentValues;
import android.content.Context;
import android.database.sqlite.SQLiteDatabase;
import android.database.sqlite.SQLiteOpenHelper;
import android.location.Location;
import android.os.Build;
import android.util.Log;
import android.util.SparseArray;

import java.io.File;
import java.lang.reflect.Array;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.List;

import at.ac.sbg.zgis.configuration.Constants;
import at.ac.sbg.zgis.configuration.Constants.Sensor;
import at.ac.sbg.zgis.configuration.Constants.SensorPlatform;
import at.ac.sbg.zgis.marker.Marker;
import at.ac.sbg.zgis.sensors.SensorDataCollection;
import at.ac.sbg.zgis.sensors.SensorDataEntry;
import at.ac.sbg.zgis.survey.Survey;
import at.ac.sbg.zgis.survey.SurveyItem;
import at.ac.sbg.zgis.survey.surveyitem.RadioButtonGroup;
import at.ac.sbg.zgis.survey.surveyitem.Slider;
import at.ac.sbg.zgis.survey.surveyitem.TextInput;
import at.ac.sbg.zgis.userinterface.eDiary;

/**
 * Direct access to database,contains all function for creating,
 * inserting and querying data.
 */
class DBHelper extends SQLiteOpenHelper {

    private static final String DATA_TABLE_NAME = "sensordata";
    private static final String DATA_COLUMN_ID = "id";
    private static final String DATA_COLUMN_TIME_MILLIS = "time_millis";
    private static final String DATA_COLUMN_TIME_ISO = "time_iso";
    private static final String DATA_COLUMN_PLATFORM_ID = "platform_id";
    private static final String DATA_COLUMN_SENSOR_ID = "sensor_id";
    private static final String DATA_COLUMN_VALUE_TEXT = "value_text";
    private static final String DATA_COLUMN_VALUE_INTEGER = "value_integer";
    private static final String DATA_COLUMN_VALUE_REAL = "value_real";
    private static final String DATA_COLUMN_VALUE_BOOLEAN = "value_boolean";
    private static final String PLATFORM_TABLE_NAME = "platform";
    private static final String PLATFORM_COLUMN_ID = "id";
    private static final String PLATFORM_COLUMN_NAME = "name";
    private static final String SENSOR_TABLE_NAME = "sensor";
    private static final String SENSOR_COLUMN_ID = "id";
    private static final String SENSOR_COLUMN_NAME = "name";
    private static final String SENSOR_COLUMN_DESC = "description";
    private static final String TAG = DBHelper.class.getSimpleName();
    private static final int DATABASE_VERSION = 1;

    // Survey
    private static final String SURVEY_TABLE_NAME = "survey";
    private static final String SURVEY_ID = "id";
    private static final String LATITUDE_COLUMN_NAME = "latitude";
    private static final String LONGITUDE_COLUMN_NAME = "longitude";

    // Column types
    private static final String COL_TYPE_INTEGER_NOT_NULL = " INTEGER NOT NULL ";
    private static final String COL_TYPE_TEXT_NOT_NULL = " TEXT NOT NULL ";
    private static final String MARKER_TABLE_NAME = "marker";
    private static final String MARKER_ID = "id";
    private static final String LOCATION_TABLE_NAME = "location";
    private static final String LOCATION_ID = "id";
    private static final String PROVIDER_COLUMN_NAME = "provider";
    private static final String HORIZONTAL_ACCURACY_COLUMN_NAME = "horizontal_accuracy";
    private static final String ALTITUDE_COLUMN_NAME = "altitude";
    private static final String VERTICAL_ACCURACY_COLUMN_NAME = "vertical_accuracy";
    private static final String BEARING_COLUMN_NAME = "bearing";
    private static final String BEARING_ACCURACY_COLUMN_NAME = "bearing_accuracy";
    private static final String SPEED_COLUMN_NAME = "speed";
    private static final String SPEED_ACCURACY_COLUMN_NAME = "speed_accuracy";


    private static SimpleDateFormat sdfIso = null;

    private final Survey survey;

    DBHelper(Context context, File dbFile, Survey survey) {
        super(context, dbFile.getAbsolutePath(), null, DATABASE_VERSION);
        sdfIso = new SimpleDateFormat(eDiary.timeFormatISOstring);
        this.survey = survey;
    }

    private static void insert(SensorDataCollection sdc, SQLiteDatabase db) {
        for (SensorDataEntry entry : sdc.getDataEntries()) {
            insert(entry, sdc.getPlatform(), sdc.getTimestamp(), db);
        }
    }

    private static void insert(SensorDataEntry entry, SensorPlatform platform, Calendar timestamp, SQLiteDatabase db) {

        ContentValues values = new ContentValues();
        values.put(DATA_COLUMN_TIME_MILLIS, timestamp.getTimeInMillis());
        values.put(DATA_COLUMN_TIME_ISO, sdfIso.format(timestamp.getTime()));
        values.put(DATA_COLUMN_PLATFORM_ID, platform.getId());
        values.put(DATA_COLUMN_SENSOR_ID, entry.getSensor().getId());

        Object value = entry.getValue();
        if (value instanceof String) {
            values.put(DATA_COLUMN_VALUE_TEXT, (String) value);
        } else if (value instanceof Double) {
            values.put(DATA_COLUMN_VALUE_REAL, (Double) value);
        } else if (value instanceof Float) {
            values.put(DATA_COLUMN_VALUE_REAL, (Float) value);
        } else if (value instanceof Byte) {
            values.put(DATA_COLUMN_VALUE_INTEGER, (Byte) value);
        } else if (value instanceof Short) {
            values.put(DATA_COLUMN_VALUE_INTEGER, (Short) value);
        } else if (value instanceof Integer) {
            values.put(DATA_COLUMN_VALUE_INTEGER, (Integer) value);
        } else if (value instanceof Long) {
            values.put(DATA_COLUMN_VALUE_INTEGER, (Long) value);
        } else if (value instanceof Boolean) {
            values.put(DATA_COLUMN_VALUE_BOOLEAN, (Boolean) value);
        } else if (value.getClass().isArray()) {
            values.put(DATA_COLUMN_VALUE_TEXT, toString(value));
        } else {
            Log.w(TAG, "Unhandled data type: " + value.getClass().getName());
            values.put(DATA_COLUMN_VALUE_TEXT, toString(value));
        }

        db.insert(DATA_TABLE_NAME, null, values);
    }

    private static String toString(Object obj) {
        return toString(obj, Constants.dataContainerSeparator);
    }

    private static String toString(Object obj, String separator) {
        if (obj.getClass().isArray()) {
            StringBuilder strb = new StringBuilder();
            int lastIdx = Array.getLength(obj) - 1;
            for (int i = 0; i <= lastIdx; i++) {
                strb.append(toString(Array.get(obj, i), Constants.separator));
                if (i < lastIdx) {
                    strb.append(separator);
                }
            }
            return strb.toString();
        } else {
            return obj.toString();
        }
    }

    /**
     * Creates a SQL command that creates a table for the survey
     *
     * @param survey a survey object
     * @return SQL string that creates a table in the database
     */
    private String createDynamicSurveyTable(Survey survey) {

        StringBuilder stringBuilder = new StringBuilder();

        // start of the SQL command
        String baseSQL = "CREATE TABLE " + SURVEY_TABLE_NAME +
                " (" +
                SURVEY_ID + "  INTEGER PRIMARY KEY AUTOINCREMENT, " +
                DATA_COLUMN_TIME_MILLIS + " INTEGER, " +
                DATA_COLUMN_TIME_ISO + " TEXT, " +
                LATITUDE_COLUMN_NAME + " REAL, " +
                LONGITUDE_COLUMN_NAME + " REAL ";
        stringBuilder.append(baseSQL);

        // Loop through survey items
        SparseArray<SurveyItem> surveyItemList = survey.getSurveyItems();
        for (int i = 0; i < surveyItemList.size(); i++) {

            SurveyItem surveyItem = surveyItemList.get(i);

            // column name
            String questionColumnName = surveyItem.getQuestionDBColumnName();

            // column type
            String columnType = "";
            if (surveyItem instanceof RadioButtonGroup) {
                columnType = COL_TYPE_TEXT_NOT_NULL;

            } else if (surveyItem instanceof Slider) {
                columnType = COL_TYPE_INTEGER_NOT_NULL;
            }

            // Append SQL snippet for creating a column for the question
            String createQuestionColumnSQL = ", " + questionColumnName + " " + columnType;
            stringBuilder.append(createQuestionColumnSQL);
        }


        String closingBracket = ")";
        stringBuilder.append(closingBracket);

        Log.d(TAG, "Created SQL for dynamic survey.");

        return stringBuilder.toString();
    }

    @Override
    public void onCreate(SQLiteDatabase db) {


        String createSensorDataTable = "CREATE TABLE " + DATA_TABLE_NAME +
                " (" +
                DATA_COLUMN_ID + " INTEGER PRIMARY KEY AUTOINCREMENT, " +
                DATA_COLUMN_TIME_MILLIS + " INTEGER, " +
                DATA_COLUMN_TIME_ISO + " TEXT, " +
                DATA_COLUMN_PLATFORM_ID + " INTEGER, " +
                DATA_COLUMN_SENSOR_ID + " INTEGER, " +
                DATA_COLUMN_VALUE_TEXT + " TEXT, " +
                DATA_COLUMN_VALUE_INTEGER + " INTEGER, " +
                DATA_COLUMN_VALUE_REAL + " REAL, " +
                DATA_COLUMN_VALUE_BOOLEAN + " INTEGER" +
                ")";

        String createPlatformTable = "CREATE TABLE " + PLATFORM_TABLE_NAME +
                " (" +
                PLATFORM_COLUMN_ID + " INTEGER PRIMARY KEY NOT NULL, " +
                PLATFORM_COLUMN_NAME + " TEXT NOT NULL" +
                ")";

        String createSensorTable = "CREATE TABLE " + SENSOR_TABLE_NAME +
                " (" +
                SENSOR_COLUMN_ID + " INTEGER PRIMARY KEY NOT NULL, " +
                SENSOR_COLUMN_NAME + " TEXT NOT NULL, " +
                SENSOR_COLUMN_DESC + " TEXT" +
                ")";

        String createMarkerTable = "CREATE TABLE " + MARKER_TABLE_NAME +
                " (" +
                MARKER_ID + "  INTEGER PRIMARY KEY AUTOINCREMENT, " +
                DATA_COLUMN_TIME_MILLIS + " INTEGER, " +
                DATA_COLUMN_TIME_ISO + " TEXT, " +
                LATITUDE_COLUMN_NAME + " REAL, " +
                LONGITUDE_COLUMN_NAME + " REAL " +
                ")";

        String createLocationTable = "CREATE TABLE " + LOCATION_TABLE_NAME +
                " (" +
                LOCATION_ID + " INTEGER PRIMARY KEY AUTOINCREMENT," +
                DATA_COLUMN_TIME_MILLIS + " INTEGER NOT NULL," +
                DATA_COLUMN_TIME_ISO + " TEXT," +
                LATITUDE_COLUMN_NAME + " REAL NOT NULL," +
                LONGITUDE_COLUMN_NAME + " REAL NOT NULL," +
                PROVIDER_COLUMN_NAME + " TEXT NOT NULL," +
                HORIZONTAL_ACCURACY_COLUMN_NAME + " REAL," +
                ALTITUDE_COLUMN_NAME + " REAL," +
                VERTICAL_ACCURACY_COLUMN_NAME + " REAL," +
                BEARING_COLUMN_NAME + " REAL," +
                BEARING_ACCURACY_COLUMN_NAME + " REAL," +
                SPEED_COLUMN_NAME + " REAL," +
                SPEED_ACCURACY_COLUMN_NAME + " REAL" +
                ")";

        db.execSQL(createSensorDataTable);
        db.execSQL(createPlatformTable);
        db.execSQL(createSensorTable);
        db.execSQL(createDynamicSurveyTable(survey));
        db.execSQL(createMarkerTable);
        db.execSQL(createLocationTable);

        fillPlatforms(db);
        fillSensors(db);
    }

    @Override
    public void onUpgrade(SQLiteDatabase db, int oldVersion, int newVersion) {
        // Nothing to do on upgrade
    }
    


    private void fillPlatforms(SQLiteDatabase db) {
        db.beginTransaction();
        try {
            for (SensorPlatform platform : SensorPlatform.values()) {
                insert(platform, db);
            }
            db.setTransactionSuccessful();
        } finally {
            db.endTransaction();
        }
    }

    private void fillSensors(SQLiteDatabase db) {
        db.beginTransaction();
        try {
            for (Sensor sensor : Sensor.values()) {
                insert(sensor, db);
            }
            db.setTransactionSuccessful();
        } finally {
            db.endTransaction();
        }
    }

    private void insert(SensorPlatform platform, SQLiteDatabase db) {
        ContentValues values = new ContentValues();
        values.put(PLATFORM_COLUMN_ID, platform.getId());
        values.put(PLATFORM_COLUMN_NAME, platform.name());
        db.insert(PLATFORM_TABLE_NAME, null, values);
    }

    private void insert(Sensor sensor, SQLiteDatabase db) {
        ContentValues values = new ContentValues();
        values.put(SENSOR_COLUMN_ID, sensor.getId());
        values.put(SENSOR_COLUMN_NAME, sensor.name());
        values.put(SENSOR_COLUMN_DESC, sensor.getDescription());
        db.insert(SENSOR_TABLE_NAME, null, values);
    }

    boolean insert(List<SensorDataCollection> sdcList) {
        SQLiteDatabase db = this.getWritableDatabase();
        db.beginTransaction();
        try {
            for (SensorDataCollection sdc : sdcList) {
                insert(sdc, db);
            }
            db.setTransactionSuccessful();
        } catch (Exception e) {
            Log.e(TAG, "Exception on insert(List<SensorDataCollection>)", e);
            return false;
        } finally {
            db.endTransaction();
            db.close();
        }
        return true;
    }

    /**
     * Inserts survey into database
     *
     * @param survey a survey object with answered surveyitems
     */
    public void insert(Survey survey) {
        ContentValues values = new ContentValues();

        // time
        Calendar timeStamp = survey.getTimestamp();
        values.put(DATA_COLUMN_TIME_MILLIS, timeStamp.getTimeInMillis());
        values.put(DATA_COLUMN_TIME_ISO, sdfIso.format(timeStamp.getTime()));

        // location
        Location location = survey.getLocation();
        values.put(LATITUDE_COLUMN_NAME, location.getLatitude());
        values.put(LONGITUDE_COLUMN_NAME, location.getLongitude());

        // survey answers
        SparseArray<SurveyItem> surveyItems = survey.getSurveyItems();
        for (int i = 0; i < surveyItems.size(); i++) {

            SurveyItem surveyItem = surveyItems.get(i);
            String questionColName = surveyItem.getQuestionDBColumnName();

            if (surveyItem instanceof Slider) {
                values.put(questionColName, ((Slider) surveyItem).getUserAnswer());
            } else if (surveyItem instanceof RadioButtonGroup) {
                values.put(questionColName, ((RadioButtonGroup) surveyItem).getUserAnswer());
            } else if (surveyItem instanceof TextInput) {
                values.put(questionColName, ((TextInput) surveyItem).getUserAnswer());
            }
            // TODO: 23.08.18 handle case properly
            // something else should not happen
        }

        SQLiteDatabase db = this.getWritableDatabase();
        db.insert(SURVEY_TABLE_NAME, null, values);

        Log.d(TAG, "Inserted survey: " + survey);
    }

    public void insert(Marker marker) {
        ContentValues values = new ContentValues();

        Calendar timeStamp = marker.getTimeStamp();
        Location location = marker.getLocation();

        values.put(DATA_COLUMN_TIME_MILLIS, timeStamp.getTimeInMillis());
        values.put(DATA_COLUMN_TIME_ISO, sdfIso.format(timeStamp.getTime()));

        values.put(LATITUDE_COLUMN_NAME, location.getLatitude());
        values.put(LONGITUDE_COLUMN_NAME, location.getLongitude());

        SQLiteDatabase db = this.getWritableDatabase();
        db.insert(MARKER_TABLE_NAME, null, values);

        Log.d(TAG, "Inserted marker");
    }

    public void insert(Location location){
        ContentValues values = new ContentValues();

        location.getTime();

        values.put(DATA_COLUMN_TIME_MILLIS, location.getTime());
        values.put(DATA_COLUMN_TIME_ISO, sdfIso.format(location.getTime()));
        values.put(LATITUDE_COLUMN_NAME, location.getLatitude());
        values.put(LONGITUDE_COLUMN_NAME, location.getLongitude());
        values.put(PROVIDER_COLUMN_NAME, location.getProvider());
        values.put(HORIZONTAL_ACCURACY_COLUMN_NAME, location.getAccuracy());
        values.put(ALTITUDE_COLUMN_NAME, location.getAltitude());
        values.put(BEARING_COLUMN_NAME, location.getBearing());
        values.put(SPEED_COLUMN_NAME, location.getSpeed());

        // some properties are only available on API 26 and higher
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
            values.put(VERTICAL_ACCURACY_COLUMN_NAME, location.getVerticalAccuracyMeters());
            values.put(BEARING_ACCURACY_COLUMN_NAME, location.getBearingAccuracyDegrees());
            values.put(SPEED_ACCURACY_COLUMN_NAME, location.getSpeedAccuracyMetersPerSecond());
        }

        SQLiteDatabase db = this.getWritableDatabase();
        db.insert(LOCATION_TABLE_NAME, null, values);


        Log.d(TAG, "Inserted location.");
    }
}
