package at.ac.sbg.zgis.recording;

import android.bluetooth.BluetoothAdapter;
import android.location.Location;
import android.util.Log;

import java.io.File;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

import at.ac.sbg.zgis.marker.Marker;
import at.ac.sbg.zgis.sensors.SensorDataBus;
import at.ac.sbg.zgis.sensors.SensorDataBusListener;
import at.ac.sbg.zgis.sensors.SensorDataCollection;
import at.ac.sbg.zgis.survey.Survey;
import at.ac.sbg.zgis.survey.SurveyItem;
import at.ac.sbg.zgis.userinterface.eDiary;
import at.ac.sbg.zgis.util.Utils;


/**
 * Singleton that has access to database
 * manages recording session
 */
public class Recording {

    private static final Recording instance = new Recording();

    private static final String DB_FILE_EXTENSION = ".sqlite";
    private static final String TAG = Recording.class.getSimpleName();
    private static final String DATABASE_DIR = "sqlite";
    private static final int cacheSize = 100;

    // BlockingQueue is used as buffer between SensorDataBus-NotificationThread and SQLiteStorage-StorageThread
    private static BlockingQueue<SensorDataCollection> queue;
    private static Thread storageThread;
    private static SensorDataBusListener dataListener;

    private static DBHelper dbHelper = null;

    // indicates if the current recording is running or if is stopped
    private static boolean running = false;

    // TODO: 07.12.18 better get location via locationManager.getLastKnownLocation()
    private Location lastLocation;

    private Survey survey;

    private Recording() {
    }

    public static Recording getInstance() {
        return instance;
    }

    private void createDatabase(Survey survey) {
        this.survey = survey;
        File dbDir = getOrCreateDirectory();
        String dbFileName = createDatabaseFileName();
        File dbFile = new File(dbDir, dbFileName);
        Log.i(TAG, "Database directory: " + dbDir.getAbsolutePath());

        dbHelper = new DBHelper(eDiary.context, dbFile, survey);
        Log.i(TAG, "Database name: " + dbHelper.getDatabaseName());
    }

    /**
     * Stops storing the data to the SQLite database and flushes the remaining items in the cache.
     * This method is blocking until all remaining items are stored.
     * <p>
     * Internally this method unregisters the listener from the SensorDataBus and interrupts the thread
     * responsible for caching and storing data.
     */
    public void stopAndFlush() {
        if (isNotRunning()) {
            Log.e(TAG, "SQLiteStorage already stopped.");
            return;
        }
        unregisterListener(dataListener);
        dataListener = null;

        storageThread.interrupt();
        try {
            storageThread.join();
        } catch (InterruptedException e) {
            Thread.currentThread().interrupt();
            Log.e(TAG, "Unexpected InterruptedException occurred.", e);
        }
        // not sure if this call is needed
        dbHelper.close();

        running = false;
    }

    private void registerListener(SensorDataBusListener dataListener) {
        SensorDataBus.subscribe(dataListener);
    }

    private void unregisterListener(SensorDataBusListener dataListener) {
        SensorDataBus.unsubscribe(dataListener);
    }

    private String createDatabaseFileName() {
        SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd'T'HHmm");
        String timestamp = sdf.format(eDiary.getTimeStamp().getTime());
        //gets the user defined nickname of the phone. Might not work on other phone models
        String deviceName = BluetoothAdapter.getDefaultAdapter().getName();
        return deviceName + "_" + timestamp + DB_FILE_EXTENSION;
    }


    /**
     * Resets the survey of the recording singleton and returns it
     */
    public void resetSurvey() {
        this.survey.resetSurvey();
    }

    /**
     * Enter answered survey
     *
     * @param index              number of surveyItem
     *
     */
    public void enterAnsweredSurveyItem(Integer index, SurveyItem answeredSurveyItem) {

        // TODO: 01.06.18 check if surveyItem is actually answered
        this.survey.getSurveyItems().put(index, answeredSurveyItem);

        Log.d(TAG, "Added answer index " + index + " surveyItem: " + answeredSurveyItem);

    }

    private File getOrCreateDirectory() {
        // This method can be used to decide whether to create an external or internal directory, based on available resources.
        return Utils.getOrCreateExternalDirectory(Recording.DATABASE_DIR);
    }

    private void saveSensorDataCollections(List<SensorDataCollection> sdcs) {
        if (dbHelper != null) {
            boolean success = dbHelper.insert(sdcs);
            Log.i(TAG, "Insertion success: " + success);
        }
    }

    public boolean isNotRunning() {
        return !running;
    }

    public void start(Survey survey) {
        createDatabase(survey);
        queue = new LinkedBlockingQueue<>();

        storageThread = new Thread(new StorageTask());
        storageThread.start();

        dataListener = new SQLiteStorageDataBusListener();
        registerListener(dataListener);
        running = true;
    }

    // TODO: 01.06.18 Check if survey is actually completely filled

    /**
     * Inserts the survey of the recoring object into the DB
     * Survey must be filled before
     */
    public void submitSurveyToDB() {
        dbHelper.insert(this.survey);
    }

    public void submitMarkerToDB(Marker marker) {
        dbHelper.insert(marker);
    }

    public void submitLocationToDB(Location location) {
        dbHelper.insert(location);
    }

    /**
     * Return not-empty survey.
     * In contrast to resetSurvey() it can contain answered SurveyItems
     */
    public Survey getSurvey() {
        return survey;
    }

    public void setLastLocation(Location lastLocation) {
        this.lastLocation = lastLocation;
    }

    public Location getLastLocation() {
        return lastLocation;
    }

    private class SQLiteStorageDataBusListener implements SensorDataBusListener {
        @Override
        public void onDataAvailable(SensorDataCollection dataCollection) {
            try {
                queue.put(dataCollection);
            } catch (Exception e) {
                Log.e(TAG, "Exception on adding element to queue.", e);
            }
        }
    }

    private class StorageTask implements Runnable {

        private final List<SensorDataCollection> cache;
        private final int cacheSize;

        StorageTask() {
            this.cache = new ArrayList<>(Recording.cacheSize);
            this.cacheSize = Recording.cacheSize;
        }

        @Override
        public void run() {

            while (!Thread.interrupted()) {
                SensorDataCollection sdc = null;
                try {
                    // queue.take is blocking the current thread unless/until an element is available.
                    // InterruptedException is thrown if the thread was interrupted during waiting.
                    sdc = queue.take();
                } catch (InterruptedException e) {
                    Log.e(TAG, "Storage thread interrupted. This could be intended to stop it!", e);
                    Thread.currentThread().interrupt();
                }
                if (sdc == null) {
                    continue;
                }
                cache.add(sdc);
                Log.v(TAG, "Cache size: " + cache.size() + " | Queue size: " + queue.size());
                // Save elements in cache if cache size is reached.
                if (cache.size() >= cacheSize) {
                    Log.d(TAG, "Clearing cache.");
                    saveSensorDataCollections(cache);
                    cache.clear();
                }
            }
            Log.d(TAG, "Flush cache on interrupt.");
            if (cache.isEmpty()) {
                return;
            }
            // Save remaining elements in cache when thread is interrupted.
            saveSensorDataCollections(cache);
            cache.clear();

        }
    }
}
