package at.ac.sbg.zgis.sensors;

import android.util.Log;

import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

/**
 * Central data bus class where all sensor data comes together.
 * From there, data is passed to other classes for visualization and storage.
 */
public class SensorDataBus {

    private static final String TAG = SensorDataBus.class.getSimpleName();

    // Using CopyOnWriteArrayList to prevent ConcurrentModificationException when adding or removing
    // listeners while iterating over the list. As listeners aren't added/removed very often the overhead
    // of this implementation can be neglected.
    private static final CopyOnWriteArrayList<SensorDataBusListener> listeners = new CopyOnWriteArrayList<>();
    private static final Map<SensorDataBusListener, ExecutorService> executors = new ConcurrentHashMap<>();

    // Private Constructor prevents instantiation of this class.
    // TODO: Change SensorDataBus from static class to Singleton?!
    private SensorDataBus() {

    }

    /**
     * Passes the incoming sensor data to all subscribed listeners.
     *
     * @param sensorDataCollection data from one of the connected sensors
     */
    public static void dataAvailable(SensorDataCollection sensorDataCollection) {
        for (SensorDataBusListener listener : listeners) {
            try {
                // get executor associated with listener
                ExecutorService executor = executors.get(listener);

                if (executor == null) {
                    Log.d(TAG, "No executor associated with listener.");
                    continue;
                }

                executor.execute(new SensorDataNotifierTask(listener, sensorDataCollection));
                printExecutorDebugInfo(executor);

            } catch (Exception e) {
                Log.d(TAG, "Exception on data notification: " + e.toString());
                e.printStackTrace();
            }
        }
    }

    /**
     * Adds the provided SensorDataBusListener to the static list of subscribers. If the handling of
     * the new data in the listener takes a long time it is recommended to redirect the handling to
     * a new thread in the listeners implementation, because all registered listeners will be notified
     * in one thread for each new data collection.
     *
     * @param listener SensorDataBusListener to be added to the list.
     */
    public static void subscribe(SensorDataBusListener listener) {
        synchronized (executors) {
            if (!executors.containsKey(listener)) {
                executors.put(listener, Executors.newSingleThreadExecutor());
            }
            listeners.addIfAbsent(listener);
        }
    }

    /**
     * Removes the provided SensorDataBusListener from the list of subscribers.
     *
     * @param listener listener to be removed from the list
     */
    public static void unsubscribe(SensorDataBusListener listener) {
        ExecutorService executor = null;

        synchronized (executors) {
            if (listener != null) {
                listeners.remove(listener);
                executor = executors.remove(listener);
            }
        }

        if (executor != null)
            shutdownExecutor(executor);
    }

    private static void shutdownExecutor(ExecutorService executor) {
        try {
            // First try to shutdown executor gracefully and wait for it.
            executor.shutdown();
            boolean isShutdown = executor.awaitTermination(1000, TimeUnit.MILLISECONDS);
            // Forcing shutdown if graceful shutdown did not succeed in time.
            if (!isShutdown) {
                executor.shutdownNow();
            }
        } catch (InterruptedException e) {
            System.err.println("Interrupted exception occurred during wait for executor shutdown.");
            Thread.currentThread().interrupt();
        } catch (Exception e) {
            System.err.println("Unhandled exception on executor shutdown.");
        }
    }

    // TODO: 09.04.2018 why is this never called?, maybe call when application shuts down

    private static void printExecutorDebugInfo(ExecutorService executor) {
        if (executor instanceof ThreadPoolExecutor) {
            final String TAG = "Executor - ";
            ThreadPoolExecutor tpe = (ThreadPoolExecutor) executor;
            Log.v(TAG, "Core Pool Size:    " + tpe.getCorePoolSize());
            Log.v(TAG, "Largest Pool Size: " + tpe.getLargestPoolSize());
            Log.v(TAG, "Maximum Pool Size: " + tpe.getMaximumPoolSize());
            Log.v(TAG, "Active Threads:    " + tpe.getActiveCount());
            Log.v(TAG, "Completed Tasks:   " + tpe.getCompletedTaskCount());
            Log.v(TAG, "Current queue size:" + tpe.getQueue().size());
        }
    }

    private static class SensorDataNotifierTask implements Runnable {
        private final SensorDataBusListener subscriber;
        private final SensorDataCollection sdc;

        SensorDataNotifierTask(SensorDataBusListener listener, SensorDataCollection sdc) {
            this.subscriber = listener;
            this.sdc = sdc;
        }

        @Override
        public void run() {
            try {
                this.subscriber.onDataAvailable(sdc);
            } catch (Exception e) {
                System.err.println("Exception occurred during notification of listener.");
                e.printStackTrace();
            }
        }
    }
}
