package at.ac.sbg.zgis.sensors;

import java.util.List;
import java.util.ArrayList;
import java.util.Calendar;

import at.ac.sbg.zgis.configuration.Constants;
import at.ac.sbg.zgis.userinterface.eDiary;

/**
 * SensorDataCollection contains a list of one or many SensorDataEntry objects along with a
 * timestamp and information about the sensor platform with which the data was collected.
 */
public class SensorDataCollection {

    private final Constants.SensorPlatform platform;
    private final Calendar timestamp;
    private final List<SensorDataEntry> dataEntries = new ArrayList<>();

    /**
     * This constructor creates a new SensorDataCollection associated with the provided SensorPlatform.
     * The timestamp of the collection is set to the current time of its creation.
     *
     * @param platform The SensorPlatform the data in the collection is belonging to.
     */
    public SensorDataCollection(Constants.SensorPlatform platform) {
        this.platform = platform;
        this.timestamp = eDiary.getTimeStamp();
    }

    /**
     * Adds new sensor data to the list in this SensorDataCollection.
     *
     * @param sensorDataEntry sensor data object containing new measurements.
     */
    public void addDataEntry(SensorDataEntry sensorDataEntry) {
        this.dataEntries.add(sensorDataEntry);
    }

    /**
     * Getter for the associated SensorPlatform of this collection.
     *
     * @return The associated SensorPlatform of the collection.
     */
    public Constants.SensorPlatform getPlatform() {
        return this.platform;
    }

    /**
     * Getter for the associated timestamp of this collection.
     *
     * @return The associated timestamp of the collection.
     */
    public Calendar getTimestamp() {
        return this.timestamp;
    }

    /**
     * Getter for the SensorDataEntries contained in the collection.
     *
     * @return The SensorDataEntries contained in the collection.
     */
    public List<SensorDataEntry> getDataEntries() {
        return this.dataEntries;
    }

    @Override
    public String toString() {
        StringBuilder strb = new StringBuilder();
        strb.append("SensorDataCollection: ").append(this.getPlatform().name()).append(" - ").append(this.getTimestamp()).append(System.lineSeparator());
        for (SensorDataEntry dataEntry : dataEntries) {
            strb.append(dataEntry.toString()).append(System.lineSeparator());
        }
        return strb.toString();
    }


}
