package at.ac.sbg.zgis.sensors.zephyr;

import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.content.Intent;
import android.util.Log;

import java.util.Set;

import at.ac.sbg.zgis.configuration.Constants;
import at.ac.sbg.zgis.sensors.SensorDataBus;
import at.ac.sbg.zgis.sensors.SensorDataCollection;
import at.ac.sbg.zgis.sensors.SensorDataEntry;
import at.ac.sbg.zgis.sensors.SensorPlatform;
import zephyr.android.BioHarnessBT.BTClient;
import zephyr.android.BioHarnessBT.ConnectListenerImpl;
import zephyr.android.BioHarnessBT.ConnectedEvent;
import zephyr.android.BioHarnessBT.PacketTypeRequest;
import zephyr.android.BioHarnessBT.ZephyrPacketArgs;
import zephyr.android.BioHarnessBT.ZephyrPacketEvent;
import zephyr.android.BioHarnessBT.ZephyrPacketListener;
import zephyr.android.BioHarnessBT.ZephyrProtocol;

/**
 * Connects to the Zephyr BioHarness 3 sensor platform and collects data
 */
public class BioHarness3 extends SensorPlatform {

    private static final String TAG = BioHarness3.class.getSimpleName();
    private static final Constants.SensorPlatform SENSOR_PLATFORM = Constants.SensorPlatform.BIOHARNESS_3;
    public final BluetoothAdapter btAdapter = BluetoothAdapter.getDefaultAdapter();
    private BluetoothDevice bioHarness;

    public BioHarness3() {
        super(BioHarness3.SENSOR_PLATFORM);
    }

    @Override
    protected void onHandleIntent(Intent intent) {
        Log.d(TAG, "onHandleIntent() invoked");
        Log.v(TAG, "Intent: " + intent);


        try {// Update connection status for GUI
            setConnectionStatus(ConnectionStatus.DISCONNECTED);
            connectToBh();
            Log.d(TAG, "BH connection started.");
        } catch (Exception e) {
            Log.e(TAG, "Error during initialization.", e);
        }
    }

    /**
     * Instantiate the necessary bluetooth objects, get the Mac address,
     * connect to the BioHarness and pass the BT device to the setter
     */
    public void connectToBh() {
        String btDeviceMacAddress = getMacAddress();
        // If device Mac Address is empty the BioHarness is not paired, start discovering devices
        if (btDeviceMacAddress.equals("")) {
            btAdapter.startDiscovery();
            setConnectionStatus(ConnectionStatus.DISCOVERING);
            Log.i(TAG, "Bluetooth device discovery started");
        } else {
            setConnectionStatus(ConnectionStatus.CONNECTING);
        }
        BTClient btClient = new BTClient(btAdapter, btDeviceMacAddress);
        ConnectionListener connectionListener = new ConnectionListener();
        btClient.addConnectedEventListener(connectionListener);
        bioHarness = btAdapter.getRemoteDevice(getMacAddress());

        if (btClient.IsConnected() && bioHarness != null) {
            setConnectionStatus(ConnectionStatus.CONNECTED);
            setDevice(bioHarness);
            btClient.start();
            Log.d(TAG, "BioHarness bluetooth connection successful");
        } else {
            Log.d(TAG, "Can't connect to BioHarness");
            onDisconnect();
        }
    }

    public void onDisconnect() {
        setConnectionStatus(ConnectionStatus.DISCONNECTED);
    }

    private String getMacAddress() {
        String bhMacID = "";

        Set<BluetoothDevice> pairedDevices = btAdapter.getBondedDevices();
        if (pairedDevices.size() > 0) {
            for (BluetoothDevice btDev : pairedDevices) {
                // TODO: 19.12.19 Jakob Miksch: maybe we can find a cleaner way to identify BH
                if (btDev.getName().toLowerCase().startsWith("bh")) {
                    bhMacID = btDev.getAddress();
                    return bhMacID;
                }
            }
        }
        return bhMacID;
    }

    public String getPairedDeviceName() {
        String deviceName = "Not paired";

        Set<BluetoothDevice> pairedDevices = btAdapter.getBondedDevices();
        if (pairedDevices.size() > 0) {
            for (BluetoothDevice btDev : pairedDevices) {
                if (btDev.getName().toLowerCase().startsWith("bh")) {
                    deviceName = btDev.getName();
                    return deviceName;
                }
            }
        }
        return deviceName;
    }

    private void setDevice(BluetoothDevice bioH) {
        this.bioHarness = bioH;
    }

    public BluetoothDevice getDevice() {
        return this.bioHarness;
    }

    private class ConnectionListener extends ConnectListenerImpl {
        final int GP_MSG_ID = 0x20;
        final int BREATHING_MSG_ID = 0x21;
        final int ECG_MSG_ID = 0x22;
        final int RtoR_MSG_ID = 0x24;
        final int ACCEL_100mg_MSG_ID = 0x2A;

        // Creating the different Objects for different types of Packets
        private final GeneralPacketInfo GPInfo = new GeneralPacketInfo();
        private final ECGPacketInfo ECGInfoPacket = new ECGPacketInfo();
        private final BreathingPacketInfo BreathingInfoPacket = new BreathingPacketInfo();
        private final RtoRPacketInfo RtoRInfoPacket = new RtoRPacketInfo();
        private final AccelerometerPacketInfo AccInfoPacket = new AccelerometerPacketInfo();

        private final PacketTypeRequest RqPacketType = new PacketTypeRequest();

        ConnectionListener() {
            super(null, null);
        }

        public void Connected(ConnectedEvent<BTClient> eventArgs) {

            // Use this object to enable or disable the different Packet types
            RqPacketType.GP_ENABLE = true;
            RqPacketType.BREATHING_ENABLE = true;
            RqPacketType.LOGGING_ENABLE = true;
            RqPacketType.ACCELEROMETER_ENABLE = true;
            RqPacketType.ECG_ENABLE = true;
            RqPacketType.RtoR_ENABLE = true;
            RqPacketType.EVENT_ENABLE = true;
            RqPacketType.SUMMARY_ENABLE = true;

            // Creates a new ZephyrProtocol object and passes it the BTComms object
            ZephyrProtocol _protocol = new ZephyrProtocol(eventArgs.getSource().getComms(), RqPacketType);

            _protocol.addZephyrPacketEventListener(new ZephyrPacketListener() {
                public void ReceivedPacket(ZephyrPacketEvent eventArgs) {

                    ZephyrPacketArgs msg = eventArgs.getPacket();

                    int MsgID = msg.getMsgID();
                    byte[] DataArray = msg.getBytes();

                    SensorDataCollection sensorDataCollection = new SensorDataCollection(BioHarness3.SENSOR_PLATFORM);

                    switch (MsgID) {
                        case GP_MSG_ID:
                            // Battery level - gives a closer reading of the battery level than the Summary package
                            byte batteryLevel = GPInfo.GetBatteryStatus(DataArray);
                            setBatteryStatus(batteryLevel);
                            // Heart Rate
                            int HRate = GPInfo.GetHeartRate(DataArray);
                            SensorDataEntry sensorDataEntry = new SensorDataEntry(Constants.Sensor.HEARTRATE, HRate);
                            sensorDataCollection.addDataEntry(sensorDataEntry);

                            // Respiration Rate
                            double RespRate = GPInfo.GetRespirationRate(DataArray);
                            sensorDataEntry = new SensorDataEntry(Constants.Sensor.RESPIRATIONRATE, RespRate);
                            sensorDataCollection.addDataEntry(sensorDataEntry);

                            // Skin Temperature
                            double SkinTempDbl = GPInfo.GetSkinTemperature(DataArray);
                            sensorDataEntry = new SensorDataEntry(Constants.Sensor.SKINTEMPERATURE, SkinTempDbl);
                            sensorDataCollection.addDataEntry(sensorDataEntry);

                            // Posture
                            int PostureInt = GPInfo.GetPosture(DataArray);
                            sensorDataEntry = new SensorDataEntry(Constants.Sensor.POSTURE, PostureInt);
                            sensorDataCollection.addDataEntry(sensorDataEntry);

                            // Peak Acceleration
                            double PeakAccDbl = GPInfo.GetPeakAcceleration(DataArray);
                            sensorDataEntry = new SensorDataEntry(Constants.Sensor.PEAKACCELERATION, PeakAccDbl);
                            sensorDataCollection.addDataEntry(sensorDataEntry);

                            // ROG Status
                            byte ROGStatus = GPInfo.GetROGStatus(DataArray);
                            sensorDataEntry = new SensorDataEntry(Constants.Sensor.ROGSTATUS, ROGStatus);
                            sensorDataCollection.addDataEntry(sensorDataEntry);

                            // GSR
                            int gsr = GPInfo.GetGSR(DataArray);
                            sensorDataEntry = new SensorDataEntry(Constants.Sensor.GSR, gsr);
                            sensorDataCollection.addDataEntry(sensorDataEntry);

                            // Breathing wave amplitude
                            double breathingWaveAmplitude = GPInfo.GetBreathingWaveAmplitude(DataArray);
                            sensorDataEntry = new SensorDataEntry(Constants.Sensor.BREATHINGWAVEAMPLITUDE, breathingWaveAmplitude);
                            sensorDataCollection.addDataEntry(sensorDataEntry);

                            // ECG amplitude
                            double ecgAmplitude = GPInfo.GetECGAmplitude(DataArray);
                            sensorDataEntry = new SensorDataEntry(Constants.Sensor.ECGAMPLITUDE, ecgAmplitude);
                            sensorDataCollection.addDataEntry(sensorDataEntry);

                            // ECG noise
                            double ecgNoise = GPInfo.GetECGNoise(DataArray);
                            sensorDataEntry = new SensorDataEntry(Constants.Sensor.ECGNOISE, ecgNoise);
                            sensorDataCollection.addDataEntry(sensorDataEntry);

                            // XYZ Acceleration Min/Max
                            double xAccnMin = GPInfo.GetX_AxisAccnMin(DataArray);
                            double xAccnMax = GPInfo.GetX_AxisAccnPeak(DataArray);
                            double yAccnMin = GPInfo.GetY_AxisAccnMin(DataArray);
                            double yAccnMax = GPInfo.GetY_AxisAccnPeak(DataArray);
                            double zAccnMin = GPInfo.GetZ_AxisAccnMin(DataArray);
                            double zAccnMax = GPInfo.GetZ_AxisAccnPeak(DataArray);
                            String xyzAccMinMax = xAccnMin + Constants.separator;
                            xyzAccMinMax += xAccnMax + Constants.separator;
                            xyzAccMinMax += yAccnMin + Constants.separator;
                            xyzAccMinMax += yAccnMax + Constants.separator;
                            xyzAccMinMax += zAccnMin + Constants.separator;
                            xyzAccMinMax += zAccnMax;

                            sensorDataEntry = new SensorDataEntry(Constants.Sensor.XYZACCELERATIONMINMAX, xyzAccMinMax);
                            sensorDataCollection.addDataEntry(sensorDataEntry);

                            // VMU Data
                            double vmu = GPInfo.GetVMU(DataArray);
                            sensorDataEntry = new SensorDataEntry(Constants.Sensor.VMU, vmu);
                            sensorDataCollection.addDataEntry(sensorDataEntry);

                            // Worn Status
                            byte wornStatus = GPInfo.GetWornStatus(DataArray);
                            sensorDataEntry = new SensorDataEntry(Constants.Sensor.WORNSTATUS, wornStatus);
                            sensorDataCollection.addDataEntry(sensorDataEntry);
                            break;

                        case BREATHING_MSG_ID:
                            // Breathing Data
                            short[] breathingSamples = BreathingInfoPacket.GetBreathingSamples(DataArray);
                            String breathingDataPacket = getStringFromDataArray(breathingSamples);
                            sensorDataEntry = new SensorDataEntry(Constants.Sensor.BREATHING, breathingDataPacket);
                            sensorDataCollection.addDataEntry(sensorDataEntry);
                            break;

                        case ECG_MSG_ID:
                            // ECG Data
                            short[] ecgSamples = ECGInfoPacket.GetECGSamples(DataArray);
                            String ecgDataPacket = getStringFromDataArray(ecgSamples);
                            sensorDataEntry = new SensorDataEntry(Constants.Sensor.ECG, ecgDataPacket);
                            sensorDataCollection.addDataEntry(sensorDataEntry);
                            break;

                        case RtoR_MSG_ID:
                            // RR Data
                            int[] rrSamples = RtoRInfoPacket.GetRtoRSamples(DataArray);
                            String rrDataPacket = getStringFromDataArray(rrSamples);
                            sensorDataEntry = new SensorDataEntry(Constants.Sensor.RRDATA, rrDataPacket);
                            sensorDataCollection.addDataEntry(sensorDataEntry);
                            break;

                        // Acceleration
                        case ACCEL_100mg_MSG_ID:
                            double[] xAxisData = AccInfoPacket.GetX_axisAccnData();
                            double[] yAxisData = AccInfoPacket.GetY_axisAccnData();
                            double[] zAxisData = AccInfoPacket.GetZ_axisAccnData();
                            String accDataPacket = getStringFromAccArray(xAxisData, yAxisData, zAxisData);
                            sensorDataEntry = new SensorDataEntry(Constants.Sensor.ACCELERATIONDATA, accDataPacket);
                            sensorDataCollection.addDataEntry(sensorDataEntry);
                            break;

                    }
                    SensorDataBus.dataAvailable(sensorDataCollection);
                }
            });
        }

        private String getStringFromAccArray(double[] x, double[] y, double[] z) {
            StringBuilder out = new StringBuilder();
            if (x.length != y.length || x.length != z.length) {
                return out.toString();
            }
            for (int i = 0; i < x.length; i++) {
                out.append(x[i]).append(Constants.separator);
                out.append(y[i]).append(Constants.separator);
                out.append(z[i]);
                if (i < x.length - 1) {
                    out.append(Constants.dataContainerSeparator);
                }
            }
            return out.toString();
        }

        private String getStringFromDataArray(int[] data_in) {
            StringBuilder out = new StringBuilder();
            for (int i = 0; i < data_in.length; i++) {
                out.append(data_in[i]);
                if (i < data_in.length - 1) {
                    out.append(Constants.dataContainerSeparator);
                }
            }
            return out.toString();
        }

        private String getStringFromDataArray(short[] data_in) {
            StringBuilder out = new StringBuilder();
            for (int i = 0; i < data_in.length; i++) {
                out.append(data_in[i]);
                if (i < data_in.length - 1) {
                    out.append(Constants.dataContainerSeparator);
                }
            }
            return out.toString();
        }
    }
}
