package at.ac.sbg.zgis.userinterface;

import android.graphics.Color;
import android.os.Bundle;
import android.os.Handler;
import android.support.v7.app.AppCompatActivity;
import android.util.Log;
import android.view.View;
import android.widget.CheckBox;

import com.github.mikephil.charting.charts.LineChart;
import com.github.mikephil.charting.components.YAxis;
import com.github.mikephil.charting.data.Entry;
import com.github.mikephil.charting.data.LineData;
import com.github.mikephil.charting.data.LineDataSet;
import com.github.mikephil.charting.interfaces.datasets.ILineDataSet;
import com.github.mikephil.charting.utils.EntryXComparator;

import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.NoSuchElementException;

import at.ac.sbg.zgis.configuration.Constants;
import at.ac.sbg.zgis.recording.Recording;
import at.ac.sbg.zgis.sensors.SensorDataBus;
import at.ac.sbg.zgis.sensors.SensorDataBusListener;
import at.ac.sbg.zgis.sensors.SensorDataCollection;
import at.ac.sbg.zgis.sensors.SensorDataEntry;

/**
 * Creates diagrams of the incoming sensor data
 */
public class RealTimeVisualization extends AppCompatActivity {

    private static final String TAG = Recording.class.getSimpleName();

    private static final int cacheSize = 100;
    private static LinkedList<SensorDataCollection> queue;

    private final SensorDataBusListener dataListener = new VisualizationDataBusListener();

    private boolean isPaused;

    private LineChart bvp;
    private LineChart gsr;
    private LineChart ibi;
    private LineChart temp;
    private LineChart ecg;

    // Arraylists containing the realtime data
    private final List<Entry> gsrEntries = new ArrayList<>();
    private final List<Entry> bvpEntries = new ArrayList<>();
    private final List<Entry> ibiEntries = new ArrayList<>();
    private final List<Entry> tempEntries = new ArrayList<>();
    private final List<Entry> ecgEntries = new ArrayList<>();

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);

        setContentView(R.layout.activity_real_time_visualization);


        // Initialize charts
        bvp = findViewById(R.id.bvp);
        gsr = findViewById(R.id.gsr);
        ibi = findViewById(R.id.ibi);
        temp = findViewById(R.id.skintemperature);
        ecg = findViewById(R.id.ecg);

        // chart properties
        assignChartProperties(bvp, 100, -100);
        assignChartProperties(gsr, 1, 0); // typical values?
        assignChartProperties(ibi, 1, 0); // typical values?
        assignChartProperties(temp, 50, 30); // typical values?
        assignChartProperties(ecg, 1, 0); // typical values?
    }

    @Override
    protected void onStart() {
        start();
        Log.i(TAG, "Visualization onStart: ");
        super.onStart();
    }

    @Override
    protected void onPause() {
        isPaused = true;
        unregisterListener(dataListener);
        Log.i(TAG, "Visualization onPause: ");
        super.onPause();
    }

    @Override
    protected void onResume() {
        isPaused = false;
        registerListener(dataListener);
        Log.i(TAG, "Visualization onResume: ");
        super.onResume();
    }

    private void start() {

        Thread visualizationThread = new Thread(new VisualizationTask());
        queue = new LinkedList<>();

        if (!isPaused) {
            visualizationThread.start();
            registerListener(dataListener);
        }


    }

    protected void onStop() {
        super.onStop();
        Log.i(TAG, "Visualization onStop: ");
        bvpEntries.clear();
        ibiEntries.clear();
        tempEntries.clear();
        gsrEntries.clear();
        ecgEntries.clear();

        bvp.clear();
        ibi.clear();
        temp.clear();
        gsr.clear();
        ecg.clear();

        queue.clear();
        unregisterListener(dataListener);
    }

    public void toggleChart(View view) {

        boolean checked = ((CheckBox) view).isChecked();

        // Check which checkbox was clicked
        switch (view.getId()) {
            case R.id.check_box_bvp:
                if (checked) {
                    bvp.setVisibility(View.VISIBLE);
                } else {
                    bvp.setVisibility(View.GONE);
                }
                break;
            case R.id.check_box_gsr:
                if (checked) {
                    gsr.setVisibility(View.VISIBLE);
                } else {
                    gsr.setVisibility(View.GONE);
                }
                break;
            case R.id.check_box_ibi:
                if (checked) {
                    ibi.setVisibility(View.VISIBLE);
                } else {
                    ibi.setVisibility(View.GONE);
                }
                break;
            case R.id.check_box_temp:
                if (checked) {
                    temp.setVisibility(View.VISIBLE);
                } else {
                    temp.setVisibility(View.GONE);
                }
                break;
            case R.id.check_box_ecg:
                if (checked) {
                    ecg.setVisibility(View.VISIBLE);
                } else {
                    ecg.setVisibility(View.GONE);
                }
                break;
        }
    }

    private void assignChartProperties(LineChart chart, float ymax, float ymin) {
        chart.setNoDataText("No data available.");
        chart.getDescription().setEnabled(false);
        chart.setDragEnabled(false);
        chart.setScaleEnabled(false);
        chart.setPinchZoom(false);
        chart.setDrawGridBackground(true);
        chart.getLegend().setEnabled(false);
        chart.setNoDataTextColor(Color.BLACK);
        chart.getXAxis().setDrawLabels(false);


        YAxis yaxis = chart.getAxisLeft();
        yaxis.setAxisMaximum(ymax);
        yaxis.setAxisMinimum(ymin);
    }

    // Adds new entries to chart and deletes old ones
    private void addEntryToChart(LineChart chart, float input, List<Entry> entries, String name, int color) {
        try {

            LineData data = chart.getData();

            if (data != null) {

                Collections.sort(entries, new EntryXComparator());

                ILineDataSet set = data.getDataSetByIndex(0);

                if (set == null) {
                    set = createLineDataSet(entries, name, color);
                    data.addDataSet(set);
                }

                data.addEntry(new Entry(set.getEntryCount(), input), 0);

                data.notifyDataChanged();
                chart.notifyDataSetChanged();
                chart.setVisibleXRangeMaximum(cacheSize - 2);
                chart.invalidate();

                // Remove the first value
                if (set.getEntryCount() == cacheSize) {

                    set.removeFirst();

                    for (Entry entry : entries)
                        entry.setX(entry.getX() - 1);
                }

            } else {
                chart.setData(new LineData());
            }
        } catch (Exception e) {
            e.printStackTrace();
            Log.e(TAG, "addEntryToChart: FAIL", e);
        }
    }

    private LineDataSet createLineDataSet(List<Entry> entries, String name, int color) {
        LineDataSet set = null;
        try {


            set = new LineDataSet(entries, name);

            set.setColor(color);
            set.setDrawCircles(false);
            set.setDrawValues(false);


        } catch (Exception e) {
            e.printStackTrace();
            Log.e(TAG, "createLineDataSet: ERROR", e);
        }
        return set;
    }

    private void registerListener(SensorDataBusListener dataListener) {
        SensorDataBus.subscribe(dataListener);
    }

    private void unregisterListener(SensorDataBusListener dataListener) {
        SensorDataBus.unsubscribe(dataListener);
    }

    private void visualize(List<SensorDataCollection> queue) {

        try {

            List<SensorDataEntry> dataEntries = queue.get(0).getDataEntries();

            Constants.Sensor sensorType = null;
            Object sensorValue = null;

            //prevents Index OOB exception
            if (dataEntries.size() != 0) {
                sensorType = dataEntries.get(0).getSensor();
                sensorValue = dataEntries.get(0).getValue();

            }

            if (sensorValue != null) {

                if (sensorType == Constants.Sensor.GSR) {
                    addEntryToChart(gsr, Float.parseFloat(sensorValue.toString()), gsrEntries, sensorType.getDescription(), getResources().getColor(R.color.red));
                }

                if (sensorType == Constants.Sensor.BVP) {
                    addEntryToChart(bvp, Float.parseFloat(sensorValue.toString()), bvpEntries, sensorType.getDescription(), getResources().getColor(R.color.blue));
                }

                if (sensorType == Constants.Sensor.IBI) {
                    addEntryToChart(ibi, Float.parseFloat(sensorValue.toString()), ibiEntries, sensorType.getDescription(), getResources().getColor(R.color.green));
                }

                if (sensorType == Constants.Sensor.SKINTEMPERATURE) {
                    addEntryToChart(temp, Float.parseFloat(sensorValue.toString()), tempEntries, sensorType.getDescription(), getResources().getColor(R.color.black));
                }

                if (sensorType == Constants.Sensor.ECG) {
                    // ECG returns a semicolon separated list of values
                    String ecgValues = sensorValue.toString();
                    // chose only the first value
                    String firstEcgValue = ecgValues.split(Constants.dataContainerSeparator)[0];

                    addEntryToChart(ecg, Float.parseFloat(firstEcgValue), ecgEntries, sensorType.getDescription(), getResources().getColor(R.color.orange));
                }
            }

        } catch (IndexOutOfBoundsException e) {

            Log.e(TAG, "visualize: INDEX OOB!!!", e);
        } catch (NegativeArraySizeException e) {
            Log.e(TAG, "visualize: NEGATIVE ARRAY SIZE!!!", e);
        } catch (Exception e) {
            Log.d(TAG, e.getMessage());
            e.printStackTrace();
        }
    }

    private class VisualizationDataBusListener implements SensorDataBusListener {
        @Override
        public void onDataAvailable(SensorDataCollection dataCollection) {

            try {
                // first in, last out
                if (queue.size() >= cacheSize)
                    queue.removeFirst();
                queue.add(dataCollection);

            } catch (NoSuchElementException e) {
                Log.e(TAG, "onDataAvailable: ", e);
            } catch (Exception e) {
                Log.e(TAG, "Exception on adding element to queue.", e);
            }
        }
    }

    private class VisualizationTask implements Runnable {

        private final int cacheSize;
        final Handler visHandler = new Handler();

        VisualizationTask() {
            this.cacheSize = RealTimeVisualization.cacheSize;
        }

        @Override
        public void run() {

            if (queue.size() >= cacheSize) {
                visualize(queue);
            }
            //Handler delays the messages sent to the visualization thread, affects the drawing speed of the visualization, set to 0 for no delay
            // The rate at which incoming data is visualized in milliseconds i.e. the delay at which the handler sends the messages to the visualization Thread
            int updateRate = 100;
            visHandler.postDelayed(this, updateRate);

        }
    }
}