#!/usr/bin/env python3

"""
Reads a SQLite database created by the eDiary app and outputs a diagram. Useful for quickly checking if something has been measured.
"""

import sqlite3
import pandas as pd
import matplotlib.pyplot as plt
import sys
import os
from geopandas import geopandas
import shapely
import contextily as ctx

# TODO: add BioHarness ECG
# TODO: add BioHarness RRdata
# TODO: add Empatica IBI

# TODO: add Basemap to map
# TODO: make map square shaped
# TODO: color location points by provider

# TODO: add sensor IDs to plot
# TODO: add metadata to plot (e.g. time, place, duration, distance, ...)

# TODO: show human-readable time
# TODO: add statistics like travel time, distance, ...


def add_basemap(ax, zoom,
                url='https://a.tile.openstreetmap.org/tileZ/tileX/tileY.png'):
    xmin, xmax, ymin, ymax = ax.axis()
    basemap, extent = ctx.bounds2img(
        xmin, ymin, xmax, ymax, zoom=zoom, url=url)
    ax.imshow(basemap, extent=extent, interpolation='bilinear')
    # restore original x/y limits
    ax.axis((xmin, xmax, ymin, ymax))


class Ediary:
    BIOHARNESS_PLATFORM_ID = 2

    def __init__(self, sqlite_path):
        self.SqlitePath = sqlite_path
        self.AllData = self.read_sensor_dataframe()
        self.E4Temp = self.get_e4_temp()
        self.E4Gsr = self.get_e4_gsr()
        self.BHHr = self.get_bh_hr()
        self.location = self.get_location()

    def read_sensor_dataframe(self):
        conn = sqlite3.connect(self.SqlitePath)
        query = "SELECT * FROM sensordata"
        # create data frames
        return pd.read_sql_query(query, conn)

    def get_location(self):
        # read from database
        conn = sqlite3.connect(self.SqlitePath)
        query = "SELECT * FROM location"
        df = pd.read_sql_query(query, conn)

        # create geodataframe
        geometry = [
            shapely.geometry.Point(x, y)
            for x, y in zip(df.longitude, df.latitude)
        ]

        gdf = geopandas.GeoDataFrame(df, geometry=geometry)
        gdf.crs = {'init': 'epsg:4326'}

        return gdf

    def get_e4_temp(self):
        df = self.AllData

        temp = df[(df.sensor_id == 3) & (df.platform_id == 3)]
        temp = temp.drop(
            columns=['platform_id', 'sensor_id',
                     'value_text', 'value_integer', 'value_boolean'])
        temp.rename(
            columns={'value_real': 'temperature_celsius'}, inplace=True)

        return temp

    def get_e4_gsr(self):
        df = self.AllData

        gsr = df[(df.sensor_id == 7) & (df.platform_id == 3)]
        gsr = gsr.drop(
            columns=['platform_id', 'sensor_id',
                     'value_text', 'value_integer', 'value_boolean'])
        gsr.rename(
            columns={'value_real': 'galvanic_skin_response'}, inplace=True)

        return gsr

    def get_bh_hr(self):
        df = self.AllData

        hr = df[(df.sensor_id == 1) & (df.platform_id == 2)]
        hr = hr.drop(
            columns=['id', 'platform_id', 'sensor_id',
                     'value_text', 'value_real',
                     'value_boolean'
                     ])
        hr.rename(
            columns={'value_integer': 'heartrate'}, inplace=True)

        hr['heartrate'] = hr['heartrate'].astype('int')

        return hr

    def axes_assign_e4_temp(self, ax):
        temp = self.E4Temp
        ax.plot(temp.time_millis, temp.temperature_celsius)
        ax.set_title('E4 - Temperature')
        return ax

    def axes_assign_e4_gsr(self, ax):
        gsr = self.E4Gsr
        ax.plot(gsr.time_millis, gsr.galvanic_skin_response)
        ax.set_title('E4 - Galvanic Skin Response')
        return ax

    def axes_assign_bh_hr(self, ax):
        hr = self.BHHr
        ax.plot(hr.time_millis, hr.heartrate)
        ax.set_title('BH - Heartrate')

    def axes_assign_location(self, ax):
        gdf = self.location
        gdf_3857 = gdf.to_crs(epsg=3857)
        gdf_3857.plot(ax=ax)
        ax.set_title('Location Trajectory')

        # basemap is very slow
        # add_basemap(ax, zoom=12)
        return ax

    def plot(self, file_path=None, show_plot=False):
        fig, axs = plt.subplots(4, 1)
        axs[0] = self.axes_assign_e4_temp(axs[0])
        axs[1] = self.axes_assign_e4_gsr(axs[1])
        axs[2] = self.axes_assign_bh_hr(axs[2])
        axs[3] = self.axes_assign_location(axs[3])
        plt.subplots_adjust(hspace=1)

        # save image to file system
        if file_path:
            if not file_path.endswith('.png'):
                file_path = file_path + '.png'
            plt.savefig(file_path)

        if (not file_path):
            show_plot = True

        # show interactive plot
        if show_plot:
            plt.show()


arguments = sys.argv
if len(arguments) == 2:

    full_file_name = arguments[1]
    filename, file_extension = os.path.splitext(full_file_name)

    ediary_sqlite = Ediary(full_file_name)
    ediary_sqlite.plot(filename + '.png')

else:
    print("Add only one positional argument with the path of the"
          "eDiary SQLite file.")
